@@ -1,14 +1,17 @@
 #if 0
 #
 # Utility for manipulating Book Type Field of Physical Format Descriptor
-# located in lead-in of DVD+RW media. This is 8th version. 2nd version
+# located in lead-in of DVD+RW media. This is 9th version. 2nd version
 # added initial support for 2nd generation DVD+RW drives. 3rd version
 # adds support for DVD+R unit settings. 4th version checks if the unit
 # is of RICOH design and reliably recognizes drive generation. 5th
 # version adds support for Benq derivatives. 6th version fixes problem
 # with USB connected units? 7th version adds support for BTC units.
 # 8th version fixes typos in BTC support, adds support for fraudulent
-# NEC firmwares and Lite-On based units.
+# NEC firmwares and Lite-On based units. 9th version adds support for
+# LG, Plextor and dual-format and double-layer Benq units. Keep in mind
+# that booktyping support might appear in certain firmware revisions,
+# not necessarily all.
 #
 # The code is in public domain.
 #
@@ -206,7 +209,7 @@ int ricoh (Scsi_Command &cmd,int action,int book,int gen=2)
   return 0;
 }
 
-int benq (Scsi_Command &cmd,int action,int book,int gen=0)
+int benq (Scsi_Command &cmd,int action,int book,int gen=2)
 { int err;
   unsigned char word[2];
   const char *brand;
@@ -250,6 +253,20 @@ int benq (Scsi_Command &cmd,int action,int book,int gen=0)
 		    printf ("The unit will brand DVD+R media as DVD%s\n",
 			    brand?brand:unknown);
 		}
+
+		if (gen<3)	break;
+
+		cmd[0]=0xFF;
+		cmd[1]=0x10;
+		cmd[2]=0x01;
+		cmd[11]=0;
+		if ((err=cmd.transport(READ,word,sizeof(word))))
+		    sperror ("BENQ_FFh(10h,1)",err);
+		else
+		{   brand = bookname (word[0],unknown),
+		    printf ("The unit will brand DVD+R DL meda as DVD%s\n",
+			    brand?brand:unknown);
+		}
 		break;
 	case OPT_MEDIA:
 		if (profile!=0x1A)
@@ -310,6 +327,19 @@ int benq (Scsi_Command &cmd,int action,int book,int gen=0)
 		else
 		    printf ("Unit was instructed to brand DVD+R as DVD%s\n",
 			    book==0x01?"-ROM":"+R");
+
+		if (gen<3)	break;
+
+		cmd[0]=0xFE;
+		cmd[0]=5;
+		cmd[2]=book==0x01?0x01:0xE1;
+		cmd[3]=1;
+		cmd[11]=0;
+		if ((err=cmd.transport()))
+		    sperror ("BENQ_FEh(5,1)",err);
+		else
+		    printf ("Unit was instructed to brand DVD+R DL as DVD%s\n",
+			    book==0x01?"-ROM":"+R DL");
 		break;
 	default:
 		break;
@@ -466,8 +496,8 @@ int liteon (Scsi_Command &cmd,int action,int book,unsigned char *buf=NULL)
 	case OPT_MEDIA:
 		fprintf (stderr,":-( Direct DVD+RW media manipulations "
 				"are not supported.\n"
-				"    Use -unit flag instead and perform "
-				"-dvd-compat recording.\n");
+				"    Use -unit flag instead and apply "
+				"-dvd-compat recordinging procedure.\n");
 		return 1;
 		break;
 	case OPT_UNIT:
@@ -496,11 +526,150 @@ int liteon (Scsi_Command &cmd,int action,int book,unsigned char *buf=NULL)
   return mediainfo (cmd);
 }
 
+// As per http://www-user.tu-chemnitz.de/~noe/Bitsetting/
+int plextor (Scsi_Command &cmd,int action,int book,int gen=2)
+{ int err;
+  unsigned char vector[8];
+
+    switch (action)
+    {	case OPT_INQ|OPT_MEDIA:
+		break;
+	case OPT_INQ|OPT_UNITRW:
+	case OPT_UNITRW:
+	case OPT_MEDIA:
+		fprintf (stderr,":-( DVD+RW booktyping is not supported\n");
+		return 1;
+		break;
+	case OPT_INQ:
+	case OPT_INQ|OPT_UNIT:
+	case OPT_INQ|OPT_UNITR:
+		cmd[0] = 0xE9;
+		cmd[2] = 0x22;
+		cmd[3] = 0x0A;
+		cmd[9] = sizeof(vector);
+		cmd[11] = 0;
+		if ((err=cmd.transport(READ,vector,sizeof(vector))))
+		    return sperror ("PLEXTOR_E9h(A)",err),1;
+
+		printf ("Unit will brand DVD+R media as DVD%s\n",
+			vector[2]?"-ROM":"+R");
+
+		if (gen<3)	return 0;
+
+		cmd[0] = 0xE9;
+		cmd[2] = 0x22;
+		cmd[3] = 0x0E;
+		cmd[9] = sizeof(vector);
+		cmd[11] = 0;
+		if ((err=cmd.transport(READ,vector,sizeof(vector))))
+		    return sperror ("PLEXTOR_E9h(E)",err),1;
+
+		printf ("Unit will brand DVD+R DL media as DVD%s\n",
+			vector[2]?"-ROM":"+R DL");
+
+		return 0;
+		break;
+	case OPT_UNIT:
+	case OPT_UNITR:
+		cmd[0] = 0xE9;
+		cmd[1] = 0x10;
+		cmd[2] = 0x22;
+		cmd[3] = 0x0A;
+		cmd[5] = book&0xF0?0:1;
+		cmd[9] = sizeof(vector);
+		cmd[11] = 0;
+		if ((err=cmd.transport(READ,vector,sizeof(vector))))
+		    return sperror ("PLEXTOR_E9h(A)",err),1;
+
+		printf ("Unit was instruction to brand DVD+R media as DVD%s\n",
+			book&0xF0?"+R":"-ROM");
+
+		if (gen<3)	return 0;
+
+		cmd[0] = 0xE9;
+		cmd[1] = 0x10;
+		cmd[2] = 0x22;
+		cmd[3] = 0x0E;
+		cmd[5] = book&0xF0?0:1;
+		cmd[9] = sizeof(vector);
+		cmd[11] = 0;
+		if ((err=cmd.transport(READ,vector,sizeof(vector))))
+		    return sperror ("PLEXTOR_E9h(E)",err),1;
+
+		printf ("Unit was instruction to brand DVD+R DL media as DVD%s\n",
+			book&0xF0?"+R DL":"-ROM");
+
+		return 0;
+		break;
+	default:
+		break;
+    }
+
+  return mediainfo (cmd);
+}
+
+int lg (Scsi_Command &cmd,int action,int book)
+{ int err;
+  unsigned char vector[4];
+
+    switch (action)
+    {	case OPT_INQ|OPT_MEDIA:
+		break;
+	case OPT_INQ|OPT_UNITRW:
+	case OPT_UNITRW:
+	case OPT_MEDIA:
+		fprintf (stderr,":-( DVD+RW booktyping is not supported\n");
+		return 1;
+		break;
+	case OPT_INQ:
+	case OPT_INQ|OPT_UNIT:
+	case OPT_INQ|OPT_UNITR:
+		cmd[0] = 0xFA;
+		cmd[8] = sizeof(vector);
+		cmd[9] = 0;
+		if ((err=cmd.transport(READ,vector,sizeof(vector))))
+		    return sperror ("LG_FAh",err),1;
+
+		printf ("Unit will brand DVD+R media as DVD%s\n",
+			vector[0]&0xF0?"+R":"-ROM");
+		printf ("Unit will brand DVD+R DL media as DVD%s\n",
+			vector[1]&0xF0?"+R DL":"-ROM");
+		return 0;
+		break;
+	case OPT_UNIT:
+	case OPT_UNITR:
+		memset (vector,0,sizeof(vector));
+		if (book&0xF0)
+		    vector[0] = 0xA0,	// DVD+R
+		    vector[1] = 0xE0;	// DVD+R DL
+
+		cmd[0] = 0xFC;
+		cmd[2] = '+';
+		cmd[3] = 'R';
+		cmd[4] = 'T';
+		cmd[5] = 'B';
+		cmd[8] = sizeof(vector);
+		cmd[9] = 0;
+		if ((err=cmd.transport(WRITE,vector,sizeof(vector))))
+		    return sperror ("LG_FCh",err),1;
+
+		printf ("Unit was instruction to brand DVD+R media as DVD%s\n",
+			book&0xF0?"+R":"-ROM");
+		return 0;
+		break;
+	default:
+		break;
+    }
+
+  return mediainfo (cmd);
+}
+
 int main(int argc,char *argv[])
 { Scsi_Command  cmd;
   unsigned char buf[128];
-  int book=0,action=0,generation=0;
-  int i,err,hp;
+  int book=0,action=0;
+  int plusgeneration=0,dashcapable=0,ramcapable=0;
+  int i,err,hp=0,plx=0;
 
     for(dev=NULL,i=1;i<argc;i++)
     {	if	(!strncmp (argv[i],"-dvd-rom",7))   book=0x01;
@@ -549,6 +718,7 @@ int main(int argc,char *argv[])
     }
     hp =       !memcmp (buf+8,"HP      ",8);
     hp = hp || !memcmp (buf+8,"_NEC    ",8);
+    plx =      !memcmp (buf+8,"PLEXTOR ",8);
 
     cmd[0]=0x46;	// GET CONFIGURATION
     cmd[1]=2;
@@ -560,7 +730,7 @@ int main(int argc,char *argv[])
     }
 
     // See if it's 2 gen drive by checking if DVD+R profile is an option.
-    // Screen out dual-format units as well...
+    // Catch dual- and triple-format units as well...
     {	int len=4+(buf[0]<<24|buf[1]<<16|buf[2]<<8|buf[3]);
 	if (len>264)
 	{   fprintf (stderr,":-( insane profile list length [%d]\n",len);
@@ -576,14 +746,16 @@ int main(int argc,char *argv[])
 	if ((err=cmd.transport(READ,list,len)))
 	    return sperror("GET CONFIGURATION",err),1;
 
-	for (generation=1,i=12;i<list[11];i+=4)
+	for (plusgeneration=1,i=12;i<list[11];i+=4)
 	{   switch (list[i]<<8|list[i+1])
 	    {	case 0x1B:	// DVD+R supported
-			if (generation) generation=2;
-			break;
+			plusgeneration=2;	break;
+		case 0x2B:	// DVD+R DL supported
+			plusgeneration=3;	break;
 		case 0x11:	// DVD-R supported
-			generation=0;
-			break;
+			dashcapable=1;		break;
+		case 0x12:	// DVD-RAM supported
+			ramcapable=1;		break;
 	    }
 	}
 
@@ -600,7 +772,7 @@ int main(int argc,char *argv[])
 	return ret;
     }
 
-    if (generation>0 || hp)	// generation==0 means dual-format unit...
+    if (!dashcapable || hp)
     {	// See if it's a RICOH design...
 	cmd[0]=0x5A;	// MODE SENSE
 	cmd[1]=0x08;	// "Disable Block Descriptors"
@@ -608,19 +780,14 @@ int main(int argc,char *argv[])
 	cmd[8]=12;
 	cmd[9]=0;
 	if (!(err=cmd.transport(READ,buf,12)) && buf[8] == 0x30)
-	    return ricoh (cmd,action,book,generation);
+	    return ricoh (cmd,action,book,plusgeneration);
 
-	// See if it's BENQ design...
-	cmd[0]=0xFD;
-	cmd[1]=0xF1;
-	cmd[2]='B';
-	cmd[3]='E';
-	cmd[4]='N';
-	cmd[5]='Q';
-	cmd[11]=0;
-	if (!(err=cmd.transport()))
-	    return benq (cmd,action,book);
+	goto benq;
     }
+    else if (plx)	// Plextor doesn't seem to re-badge their units...
+	return plextor (cmd,action,book,plusgeneration);
+    else if (ramcapable) // LG is the only one we know...
+	return lg (cmd,action,book);
     else
     {	// See if it's BTC design...
 	cmd[0]=0xFA;
@@ -641,10 +808,22 @@ int main(int argc,char *argv[])
 	cmd[11]=0;
 	if (!(err=cmd.transport(READ,buf,sizeof(buf))))
 	    return liteon (cmd,action,book,buf);
+
+    benq:
+	// See if it's BENQ design...
+	cmd[0]=0xFD;
+	cmd[1]=0xF1;
+	cmd[2]='B';
+	cmd[3]='E';
+	cmd[4]='N';
+	cmd[5]='Q';
+	cmd[11]=0;
+	if (!(err=cmd.transport()))
+	    return benq (cmd,action,book,plusgeneration);
     }
 
     fprintf (stderr,"This program targets units of "
-		    "RICOH, BENQ, BTC and LITE-ON designs.\n"
+		    "RICOH, BENQ, BTC, LITE-ON, LG and PLEXTOR designs.\n"
 		    "%s doesn't appear to be one. Exiting.\n",
 		    dev);
 
