@@ -1,5 +1,5 @@
 /*
- * DVDRW format 4.10 by Andy Polyakov <appro@fy.chalmers.se>.
+ * DVDRW format 6.1 by Andy Polyakov <appro@fy.chalmers.se>.
  *
  * Use-it-on-your-own-risk, GPL bless...
  *
@@ -55,10 +55,20 @@
  * - permit for DVD-RW blank even if format descriptors are not present;
  * 4.10:
  * - add support for DVD-RAM;
+ * 6.0:
+ * - versioning harmonization;
+ * - WIN32 port;
+ * - Initial DVD+RW Double Layer support;
+ * - Ignore "WRITE PROTECTED" error in OPC;
+ * 6.1:
+ * -  localization;
+ * - Treat only x73xx OPC errors as fatal;
  */
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+
+#if defined(__unix) || defined(__unix__)
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -67,22 +77,56 @@
 #include <signal.h>
 #include <setjmp.h>
 #include <sys/wait.h>
-#include <errno.h>
+#endif
 
 #include "transport.hxx"
 
 static void usage (char *prog)
 { fprintf (stderr,"- usage: %s [-force[=full]] [-lead-out|-blank[=full]]\n"
-		  "         [-ssa[=none|default|max]] /dev/cdrom\n",prog),
+		  "         [-ssa[=none|default|max]] /dev/dvd\n",prog),
   exit(1);
 }
 
-static volatile int *progress;
-static const char   *gui_action=NULL;
+#ifdef _WIN32
+#include <process.h>
 
+# if defined(__GNUC__)
+#  define __shared__	__attribute__((section(".shared"),shared))
+# elif defined(_MSC_VER)
+#  define __shared__
+#  pragma data_seg(".shared")
+#  pragma comment(linker,"/section:.shared,rws")
+# endif
+static volatile int shared_progress_indicator __shared__ = 0;
+# if defined(_MSC_VER)
+#  pragma data_seg()
+# endif
+
+static volatile int *progress = &shared_progress_indicator;
+static HANDLE Process    = NULL;
+
+BOOL WINAPI GoBackground (DWORD code)
+{   if (*progress)
+    {	FreeConsole();	/* detach from console upon ctrl-c or window close */
+	return TRUE;
+    }
+  return FALSE;
+}
+
+void KillForeground (void)
+{   fprintf(stderr,"\n");
+    TerminateProcess(Process,0);
+}
+#else
+static volatile int *progress;
+#endif
+
+static const char   *gui_action=NULL;
+static const char   *str = "|/-\\",*backspaces="\b\b\b\b\b\b\b\b\b\b";
+
+#if defined(__unix) || defined(__unix__)
 extern "C" void alarm_handler (int no)
 { static int         i=0,len=1,old_progress=0;
-  static const char *str = "|/-\\",*backspaces="\b\b\b\b\b\b\b\b\b\b";
   int new_progress = *progress;
 
     if (gui_action)
@@ -102,6 +146,7 @@ extern "C" void alarm_handler (int no)
 
     alarm(1);
 }
+#endif
 
 int main (int argc, char *argv[])
 { Scsi_Command	cmd;
@@ -109,10 +154,64 @@ int main (int argc, char *argv[])
   char		*dev=NULL,*p;
   unsigned int	capacity,lead_out,mmc_profile,err;
   int		force=0,full=0,compat=0,blank=0,ssa=0,do_opc=0,gui=0,len,i;
+#if defined(__unix) || defined(__unix__)
   pid_t		pid;
+#endif
 
-    fprintf (stderr,"* DVDRW/-RAM format utility by <appro@fy.chalmers.se>, "
-		    "version 4.10.\n");
+#ifdef _WIN32
+  DWORD		ppid;
+  char		filename[MAX_PATH],*progname;
+
+    /*
+     * Foreground process spawns itself and simply waits for shared
+     * progress indicator to increment...
+     */
+    if (sscanf(argv[0],":%u:",&ppid) != 1)
+    { int  i=0,len=1,old_progress=0,new_progress;
+
+	sprintf (filename,":%u:",GetCurrentProcessId());
+	progname = argv[0];
+	argv[0]  = filename;
+	Process  = (HANDLE)_spawnv (_P_NOWAIT,progname,argv);
+	if (Process == (HANDLE)-1)
+	    perror("_spawnv"), ExitProcess(errno);
+
+	while(1)
+	{   if (WaitForSingleObject(Process,1000) == WAIT_OBJECT_0)
+	    {	ppid = 0; /* borrow DWORD variable */
+		GetExitCodeProcess(Process,&ppid);
+		ExitProcess(ppid);
+	    }
+
+	    new_progress = *progress;
+	    if (new_progress != old_progress)
+		len = fprintf (stderr,"%.*s%.1f%%",len,backspaces,
+					100.0*new_progress/65536.0) - len,
+		old_progress = new_progress;
+	    else if (new_progress)
+		fprintf (stderr,"\b%c",str[i]),
+		i++, i&=3;
+	}
+    }
+
+    /*
+     * ... while background process does *all* the job...
+     */
+    Process = OpenProcess (PROCESS_TERMINATE,FALSE,ppid);
+    if (Process == NULL)
+	perror("OpenProcess"), ExitProcess(errno);
+
+    atexit (KillForeground);
+    SetConsoleCtrlHandler (GoBackground,TRUE);
+
+    GetModuleFileName (NULL,filename,sizeof(filename));
+    progname = strrchr(filename,'\\');
+    if (progname)	argv[0] = progname+1;
+    else		argv[0] = filename;
+#endif
+
+    fprintf (stderr,"* DVD%sRW/-RAM format utility by <appro@fy.chalmers.se>, "
+		    "version 6.1.\n",plusminus_locale());
 
     for (i=1;i<argc;i++) {
 	if (*argv[i] == '-')
@@ -138,7 +237,11 @@ int main (int argc, char *argv[])
 	    }
 	    else if (argv[i][1] == 'g')	gui=1;
 	    else			usage(argv[0]);
+#ifdef _WIN32
+	else if (argv[i][1] == ':')
+#else
 	else if (*argv[i] == '/')
+#endif
 	    dev = argv[i];
 	else
 	    usage (argv[0]);
@@ -168,11 +271,11 @@ int main (int argc, char *argv[])
 
     mmc_profile = formats[6]<<8|formats[7];
 
-    if (mmc_profile!=0x1A
+    if (mmc_profile!=0x1A && mmc_profile!=0x2A
 	&& mmc_profile!=0x14 && mmc_profile!=0x13
 	&& mmc_profile!=0x12)
 	fprintf (stderr,":-( mounted media doesn't appear to be "
-			"DVDRW or DVD-RAM\n"),
+			"DVD%sRW or DVD-RAM\n",plusminus_locale()),
 	exit (1);
 
     /*
@@ -213,7 +316,7 @@ int main (int argc, char *argv[])
 	fprintf (stderr,":-( parameter length inconsistency\n"),
 	exit(1);
 
-    if (mmc_profile==0x1A)	// DVD+RW
+    if (mmc_profile==0x1A || mmc_profile==0x2A)	// DVD+RW
     {	for (i=8;i<len;i+=8)	// look for "DVD+RW Full" descriptor
 	    if ((formats [4+i+4]>>2) == 0x26) break;
     }
@@ -275,7 +378,7 @@ int main (int argc, char *argv[])
     capacity |= formats[4+i+2], capacity <<= 8;
     capacity |= formats[4+i+3];
 
-    if (mmc_profile==0x1A)	// DVD+RW
+    if (mmc_profile==0x1A || mmc_profile==0x2A)	// DVD+RW
 	fprintf (stderr,"* %.1fGB DVD+RW media detected.\n",
 			2048.0*capacity/1e9);
     else if (mmc_profile==0x12)	// DVD-RAM
@@ -308,14 +411,14 @@ int main (int argc, char *argv[])
 
     if (dinfo[2]&3)		// non-blank media
     {	if (!force)
-	{   if (mmc_profile==0x1A || mmc_profile==0x13 || mmc_profile==0x12)
+	{   if (mmc_profile==0x1A || mmc_profile==0x2A || mmc_profile==0x13 || mmc_profile==0x12)
 		fprintf (stderr,"- media is already formatted, lead-out is currently at\n"
 				"  %d KiB which is %.1f%% of total capacity.\n",
 				lead_out*2,(100.0*lead_out)/capacity);
 	    else
 		fprintf (stderr,"- media is not blank\n");
 	offer_options:
-	    if (mmc_profile == 0x1A)
+	    if (mmc_profile == 0x1A || mmc_profile == 0x2A)
 		fprintf (stderr,"- you have the option to re-run %s with:\n"
 				"  -lead-out  to elicit lead-out relocation for better\n"
 				"             DVD-ROM compatibility, data is not affected;\n"
@@ -354,17 +457,18 @@ int main (int argc, char *argv[])
 	force = 0;
     }
 
-    if ((mmc_profile == 0x1A && blank)
+    if (((mmc_profile == 0x1A || mmc_profile == 0x2A) && blank)
 	|| (mmc_profile != 0x1A && compat)
 	|| (mmc_profile != 0x12 && ssa) )
     {	fprintf (stderr,"- illegal command-line option for this media.\n");
 	goto offer_options;
     }
-    else if (mmc_profile == 0x1A && full)
+    else if ((mmc_profile == 0x1A || mmc_profile == 0x2A) && full)
     {	fprintf (stderr,"- unimplemented command-line option for this media.\n");
 	goto offer_options;
     }
 
+#if defined(__unix) || defined(__unix__)
     { int fd;
       char *s;
 
@@ -418,6 +522,15 @@ int main (int argc, char *argv[])
     signal(SIGHUP,SIG_IGN);
     signal(SIGINT,SIG_IGN);
     signal(SIGTERM,SIG_IGN);
+#else
+    if (compat && force)	str="relocating lead-out";
+    else if (blank)		str="blanking";
+    else			str="formatting";
+    if (gui)			gui_action=str;
+    else			fprintf (stderr,"* %s .",str);
+#endif
+
+    *progress = 1;
 
     // formats[i] becomes "Format Unit Parameter List"
     formats[i+0] = 0;		// "Reserved"
@@ -427,7 +540,7 @@ int main (int argc, char *argv[])
 
     handle_events(cmd);
 
-    if (mmc_profile==0x1A)	// DVD+RW
+    if (mmc_profile==0x1A || mmc_profile==0x2A)	// DVD+RW
     {	if (compat && force && (dinfo[2]&3))
 	    formats[i+4+0]=formats[i+4+1]=formats[i+4+2]=formats[i+4+3]=0,
 	    formats[i+4+7] = 1;	// "Restart format"
@@ -486,7 +599,16 @@ int main (int argc, char *argv[])
 	    cmd[9] = 0;
 	    cmd.timeout(120);	// NEC units can be slooo...w
 	    if ((err=cmd.transport()))
-		sperror ("PERFORM OPC",err), exit (1);
+	    {	if (err==0x17301)	// "POWER CALIBRATION AREA ALMOST FULL"
+		    fprintf (stderr,":-! WARNING: Power Calibration Area "
+					"is almost full\n");
+		else
+		{   sperror ("PERFORM OPC",err);
+		    if ((err&0x0FF00)==0x07300)	exit (1);
+		    /* The rest of errors are ignored, see groisofs_mmc.cpp
+		     * for further details... */
+		}
+	    }
 	}
 
 	if (blank)		// DVD-RW blanking procedure
@@ -516,17 +638,7 @@ int main (int argc, char *argv[])
 	}
     }
 
-    while (1)	// Pioneer DVR-x06 needs this...
-    {	cmd[0] = 0x1B;		// START/STOP UNIT
-	cmd[1] = 0x1;		// "IMMED"
-	cmd[4] = 0;		// "Stop"
-	cmd[5] = 0;
-	if (cmd.transport() == 0x20407)	// "OP IN PROGRESS"
-	{   poll (NULL,0,333);
-	    continue;
-	}
-	break;
-    }
+    pioneer_stop (cmd,progress);
 
 #if 0
     cmd[0] = 0x1E;	// ALLOW MEDIA REMOVAL
