@@ -11,9 +11,15 @@
 
 #include "transport.hxx"
 
+#ifdef _WIN32
+#define LLU "I64u"
+#else
+#define LLU "llu"
+#endif
+
 int main(int argc,char *argv[])
 { Scsi_Command	cmd;
-  unsigned char	header[40],inq[128],*page2A=NULL;
+  unsigned char	header[48],inq[128],*page2A=NULL;
   char		cmdname[64];
   int		i,j,ntracks,err,dvd_dash=0,dvd_plus=0,
 		plus_mediaid_printed=0,page2A_printed=0,dvd_ram_spare=0;
@@ -21,23 +27,23 @@ int main(int argc,char *argv[])
 
     if (argc<2)
 	fprintf (stderr,"usage: %s /dev/dvd\n",argv[0]),
-	exit (1);
+	exit (FATAL_START(EINVAL));
 
     if (!cmd.associate(argv[1]))
 	fprintf (stderr,"%s: unable to open: ",argv[1]),
 	perror (NULL),
-	exit (1);
+	exit (FATAL_START(errno));
 
     cmd[0] = 0x12;	// INQUIRY
     cmd[4] = 36;
     cmd[5] = 0;
     if ((err=cmd.transport(READ,inq,36)))
 	sperror ("INQUIRY",err),
-	exit (1);
+	exit (FATAL_START(errno));
 
     if ((inq[0]&0x1F) != 5)
 	fprintf (stderr,":-( not an MMC unit!\n"),
-	exit (1);
+	exit (FATAL_START(EINVAL));
 
     if (argc>1)
 	printf ("INQUIRY:                [%.8s][%.16s][%.4s]\n",
@@ -51,7 +57,7 @@ int main(int argc,char *argv[])
     cmd[5] = 0;
     if ((err=cmd.transport()) == -1)
 	sperror ("TEST UNIT READY",err),
-	exit (1);
+	exit (FATAL_START(errno));
 #endif
 
     if (argc>2) do
@@ -69,11 +75,11 @@ int main(int argc,char *argv[])
 	if ((err=cmd.transport(READ,header,12)))
 	    sprintf (cmdname,"MODE SENSE#%02X",pagecode),
 	    sperror (cmdname,err),
-	    exit(1);
+	    exit(errno);
 
 	if (header[6]<<8|header[7])
 	    fprintf (stderr,":-( \"DBD\" is not respected\n"),
-	    exit(1);
+	    exit(errno);
 
 	len=(header[0]<<8|header[1])+2;
 
@@ -136,7 +142,7 @@ int main(int argc,char *argv[])
 	    cmd[9] = 0;
 	    if ((err=cmd.transport(READ,page2A,len)))
 		sperror ("MODE SENSE#2A",err),
-		exit (1);
+		exit (errno);
 
 	    if (page2A[6]<<8|page2A[7])
 		fprintf (stderr,":-( \"DBD\" is not respected\n");
@@ -150,7 +156,7 @@ int main(int argc,char *argv[])
 
 	    if (!(page2A[8+(page2A[6]<<8|page2A[7])+2]&8))
 		fprintf (stderr,":-( not a DVD unit\n"),
-		exit (1);
+		exit (FATAL_START(EINVAL));
 
 	    if (page2A==header) page2A=NULL;
 	    else if (argc>2 && !page2A_printed)
@@ -220,15 +226,13 @@ int main(int argc,char *argv[])
     {	if (err!=0x52000)	// "IVALID COMMAND OPERATION CODE"
 	    sperror ("GET CONFIGURATION",err),
 	    perror (NULL),
-	    exit(1);
+	    exit(FATAL_START(errno));
 	profile=0x10;	// non MMC3 DVD-ROM unit?
 	goto legacy;
     }
 
     profile=header[6]<<8|header[7];
 
-    printf ("GET [CURRENT] CONFIGURATION:\n");
-
     if (argc>2)
     { unsigned char *profiles;
       int len,n;
@@ -246,6 +250,8 @@ int main(int argc,char *argv[])
 	cmd[9] = 0;
 	cmd.transport(READ,profiles,len);
 
+	printf ("GET [CURRENT] CONFIGURATION:\n");
+
 	for(i=8;i<len;)
 	{   printf (" %04X%c",profiles[i]<<8|profiles[i+1],':');
 	    n=profiles[i+3]; i+=4;
@@ -261,10 +267,10 @@ int main(int argc,char *argv[])
 
     if (profile==0)
 	fprintf (stderr,":-( no media mounted, exiting...\n"),
-	exit(1);
-    if ((profile&0xF0) != 0x10 && profile != 0x2B)
+	exit(FATAL_START(ENOMEDIUM));
+    if ((profile&0xF0) != 0x10 && (profile&0xF0) != 0x20)
 	fprintf (stderr,":-( non-DVD media mounted, exiting...\n"),
-	exit (1);
+	exit (FATAL_START(EMEDIUMTYPE));
 
     { const char *str;
 
@@ -274,11 +280,17 @@ int main(int argc,char *argv[])
 	    case 0x12:	str="-RAM";			break;
 	    case 0x13:	str="-RW Restricted Overwrite";	break;
 	    case 0x14:	str="-RW Sequential";		break;
+	    case 0x15:	str="-R Dual Layer Sequential";	break;
+	    case 0x16:	str="-R Dual Layer Jump";	break;
 	    case 0x1A:	str="+RW";			break;
 	    case 0x1B:	str="+R";			break;
+	    case 0x2A:	str="+RW Double Layer";		break;
 	    case 0x2B:	str="+R Double Layer";		break;
 	    default:	str="unknown";			break;
 	}
+
+	if (argc<=2)
+	    printf ("GET [CURRENT] CONFIGURATION:\n");
 	printf (" Mounted Media:         %Xh, DVD%s\n",profile,str);
 	if (str[0]=='-' && str[3]!='M')	dvd_dash=0x10;
 	if (str[0]=='+')		dvd_plus=0x11;
@@ -287,7 +299,7 @@ int main(int argc,char *argv[])
     if (dvd_0E || dvd_11) do
     { union { unsigned char _e[4+40],_11[4+256]; } dvd;
       unsigned int  len;
-      unsigned char format=dvd_dash?0x0E:0x11;
+      unsigned char format=dvd_plus?0x11:0x0E;
 
 	cmd[0] = 0xAD;
 	cmd[7] = format;
@@ -296,7 +308,7 @@ int main(int argc,char *argv[])
 	if ((err=cmd.transport(READ,header,4)))
 	{   if (argc>2)
 		sprintf (cmdname,"READ DVD STRUCTURE#%02x",format),
-		sperror (cmdname,err);
+		sperror (cmdname,cmd.sense());
 	    break;
 	}
 
@@ -318,7 +330,7 @@ int main(int argc,char *argv[])
 	    printf("\n");
 	}
 
-	if (dvd_dash && dvd_0E && dvd._e[4+16]==3 && dvd._e[4+24]==4)
+	if (!dvd_plus && dvd_0E && dvd._e[4+16]==3 && dvd._e[4+24]==4)
 	printf (" Media ID:              %6.6s%-6.6s\n",dvd._e+4+17,dvd._e+4+25);
 
 	if (dvd_plus && dvd_11)
@@ -505,7 +517,7 @@ legacy:
     {	if (dvd_dash) { dvd_dash=0; goto legacy; }
 	if (err!=0x52400 || argc>2)
 	    sprintf (cmdname,"READ DVD STRUCTURE#%X",dvd_dash),
-	    sperror (cmdname,err);
+	    sperror (cmdname,cmd.sense());
     }
     else do
     { unsigned char book=header[4];
@@ -551,7 +563,7 @@ legacy:
 	if (phys_end>0)	phys_end -= phys_start;
 	if (phys_end>0) phys_end += 1;
 
-	printf (" %s    %u*2KB=%llu\n",
+	printf (" %s    %u*2KB=%"LLU"\n",
 		dvd_dash>=0?"Legacy lead-out at:":"Last border-out at:",
 		phys_end,phys_end*2048LL);
 
@@ -579,12 +591,12 @@ legacy:
 	cmd[9] = sizeof(s);
 	cmd[11] = 0;
 	if ((err=cmd.transport(READ,s,sizeof(s))))
-	{   sperror ("READ DVD+R DL BOUNDARY INFORMATION",err);
+	{   sperror ("READ LAYER BOUNDARY INFORMATION",err);
 	    break;
 	}
 
 	if ((s[0]<<8|s[1]) < 10)
-	{   fprintf (stderr,":-( insane DVD+R DL BI structure length\n");
+	{   fprintf (stderr,":-( insane LBI structure length\n");
 	    break;
 	}
 
@@ -594,6 +606,70 @@ legacy:
 					 s[4]&0x80?"no longer":"still");
     } while (0);
 
+    if (profile==0x16) do
+    { unsigned char s[4+8];
+
+	// Layer Jump specific structures. Note that growisofs doesn't
+	// perform Layer Jump recordings, see web-page for details...
+
+	cmd[0] = 0xAD;	// READ DVD STRUCTURE
+	cmd[7] = 0x21;	// "DVD-R DL Shifted Middle Area Start Address"
+	cmd[9] = sizeof(s);
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,s,sizeof(s))))
+	    argc>2 ? sperror ("READ SHIFTED MIDDLE AREA START ADDRESS",err) : (void)0;
+	else if ((s[0]<<8|s[1]) < 10)
+	    fprintf (stderr,":-( insane SMASA structure length\n");
+	else {
+	printf ("DVD-R DL SHIFTED MIDDLE AREA START ADDRESS:\n");
+	printf (" Shifted Middle Area:   %u*2KB, can %s be set\n",
+					 s[8]<<24|s[9]<<16|s[10]<<8|s[11],
+					 s[4]&0x80?"no longer":"still");
+	}
+
+	cmd[0] = 0xAD;	// READ DVD STRUCTURE
+	cmd[7] = 0x22;	// "DVD-R DL Jump Interval Size"
+	cmd[9] = sizeof(s);
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,s,sizeof(s))))
+	    argc>2 ? sperror ("READ JUMP INTERVAL SIZE",err) : (void)0;
+	else if ((s[0]<<8|s[1]) < 10)
+	    fprintf (stderr,":-( insane JIS structure length\n");
+	else {
+	printf ("DVD-R DL JUMP INTERVAL SIZE:\n");
+	printf (" Jump Interval Size:    %u*2KB\n",
+					 s[8]<<24|s[9]<<16|s[10]<<8|s[11]);
+	}
+
+	cmd[0] = 0xAD;	// READ DVD STRUCTURE
+	cmd[7] = 0x23;	// "DVD-R DL Manual Layer Jump Address"
+	cmd[9] = sizeof(s);
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,s,sizeof(s))))
+	    argc>2 ? sperror ("READ MANUAL LAYER JUMP ADDRESS",err) : (void)0;
+	else if ((s[0]<<8|s[1]) < 10)
+	    fprintf (stderr,":-( insane MLJA structure length\n");
+	else {
+	printf ("DVD-R DL MANUAL LAYER JUMP ADDRESS:\n");
+	printf (" Manual Layer Jump:     %u*2KB\n",
+					 s[8]<<24|s[9]<<16|s[10]<<8|s[11]);
+	}
+
+	cmd[0] = 0xAD;	// READ DVD STRUCTURE
+	cmd[7] = 0x24;	// "DVD-R DL Remapping Address"
+	cmd[9] = sizeof(s);
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,s,sizeof(s))))
+	    argc>2 ? sperror ("READ REMAPPING ADDRESS",err) : (void)0;
+	else if ((s[0]<<8|s[1]) < 10)
+	    fprintf (stderr,":-( insane RA structure length\n");
+	else {
+	printf ("DVD-R DL REMAPPING ADDRESS:\n");
+	printf (" Remapping Jump:        %u*2KB\n",
+					 s[8]<<24|s[9]<<16|s[10]<<8|s[11]);
+	}
+    } while (0);
+
     if (profile==0x12 && dvd_0A) do
     { unsigned char s[16];
       unsigned int  a,b;
@@ -649,8 +725,8 @@ legacy:
     cmd[8] = 32;
     cmd[9] = 0;
     if ((err=cmd.transport(READ,header,32)))
-	sperror ("READ DISC INFORMATION",err),
-	exit (1);
+	sperror ("READ DISC INFORMATION",cmd.sense()),
+	exit (errno);
 
     { static const char
 		   *session_state[]={	"empty",	"incomplete",
@@ -723,14 +799,14 @@ legacy:
 	    printf(" %s:\t\t%u*%u=",type[formats[8]&3],
 	        capacity=formats[4]<<24|formats[5]<<16|formats[6]<<8|formats[7],
 		blocksize=formats[9]<<16|formats[10]<<8|formats[11]);
-	    printf("%llu\n",(unsigned long long)capacity*blocksize);
+	    printf("%"LLU"\n",(unsigned long long)capacity*blocksize);
 
 	    for(i=12;i<len;i+=8)
 	    {	printf(" %02Xh(%x):\t\t%u*%u=",formats[i+4]>>2,
 			formats[i+5]<<16|formats[i+6]<<8|formats[i+7],
 	    		capacity=formats[i]<<24|formats[i+1]<<16|formats[i+2]<<8|formats[i+3],
 			blocksize);
-		printf("%llu\n",(unsigned long long)capacity*blocksize);
+		printf("%"LLU"\n",(unsigned long long)capacity*blocksize);
 	    }
 	    break;
 	}
@@ -761,7 +837,7 @@ legacy:
 	    if ((err=cmd.transport(READ,header,len)))
 	    {	if (i<ntracks) { i=ntracks; continue; }
 		sperror ("READ TRACK INFORMATION",err),
-		exit (1);
+		exit (errno);
 	    }
 
 	    if ((profile==0x1B || profile==0x2B) &&
@@ -769,6 +845,11 @@ legacy:
 	    {	if (len==32)	{ len=40; continue; }
 		else		{ break;            }
 	    }
+	    else if ((profile==0x15 || profile==0x16) &&
+		(header[5]&0xC0) && (header[0]<<8|header[1])>=46)
+	    {	if (len==32)	{ len=48; continue; }
+		else		{ break;	    }
+	    }
 
 	    len = 0;
 	}
@@ -800,7 +881,9 @@ legacy:
 	if (header[6]&0x10)
 	printf (" Fixed Packet Size:     %u*2KB\n",
 		header[20]<<24|header[21]<<16|header[22]<<8|header[23]);
-	printf (" Track Size:            %u*2KB\n",
+	printf (header[5]&0x40?		// check upon LJ bit
+		" Zone End Address:      %u*2KB\n":
+		" Track Size:            %u*2KB\n",
 		header[24]<<24|header[25]<<16|header[26]<<8|header[27]);
 	if (header[7]&2)
 	printf (" Last Recorded Address: %u*2KB\n",
@@ -808,6 +891,12 @@ legacy:
 	if (len>=40)
 	printf (" ROM Compatibility LBA: %u\n",
 		header[36]<<24|header[37]<<16|header[38]<<8|header[39]);
+	if (len>=48 && header[5]&0xC0)	// check upon LJRS bits
+	printf (" LJRS field:            %u\n",header[5]>>6),
+	printf (" Next Layer Jump:       %u\n",
+		header[40]<<24|header[41]<<16|header[42]<<8|header[43]),
+	printf (" Last Layer Jump:       %u\n",
+		header[44]<<24|header[45]<<16|header[46]<<8|header[47]);
     }
 
     do
@@ -886,11 +975,14 @@ legacy:
 	    break;
 	}
 
-	ccity = header[0]<<24|header[1]<<16|header[2]<<8|header[3], ccity++;
+	ccity = header[0]<<24|header[1]<<16|header[2]<<8|header[3];
+	if (ccity) ccity++;
 	bsize = header[4]<<24|header[5]<<16|header[6]<<8|header[7];
 
-	printf ("READ CAPACITY:          %u*%u=%llu\n",
+	printf ("READ CAPACITY:          %u*%u=%"LLU"\n",
 					ccity,bsize,
 					(unsigned long long)ccity*bsize);
     } while (0);
+
+  return 0;
 }
