@@ -18,21 +18,28 @@
 
 #include <time.h>
 #include <sys/types.h>
-#include <signal.h>
-#define sigs_mask(how)	do {	\
-    sigset_t mask;		\
-				\
-    sigemptyset (&mask);	\
-    sigaddset (&mask,SIGHUP),	sigaddset (&mask,SIGINT),	\
-    sigaddset (&mask,SIGTERM),	sigaddset (&mask,SIGPIPE);	\
-					\
-    sigprocmask (how,&mask,NULL);	\
-} while (0)
+
+#define SIGS_BLOCK	1
+#define SIGS_UNBLOCK	0
+extern "C" void sigs_mask (int);
+extern "C" void atsignals (void (*)(void));
 
 #define	ONEX 1352	// 1385 * 1000 / 1024
 
-static int           media_written=0,next_track=1,velocity=0,
-		     is_dao=0,quickgrown=0,do_reload=1;
+static int		media_written=0,next_track=1,
+			is_dao=0,quickgrown=0,do_reload=1;
+static unsigned int	stdnap=(256*1000)/ONEX,	// measured in milliseconds
+			buf_size=512;		// measured in KBs
+static class __velocity__ {
+    private:
+	int value;
+    public:
+	int operator=(int v)	{ if (v>=ONEX)	stdnap=(buf_size*500)/v;
+				  else		stdnap=(buf_size*500)/ONEX;
+				  return value = v;
+				}
+	operator int()		{ return value; }
+} velocity;
 
 static void         *ioctl_handle=(void *)-1;
 #ifndef ioctl_fd
@@ -40,12 +47,15 @@ static void         *ioctl_handle=(void *)-1;
 #endif
 
 static unsigned int  next_wr_addr=1;	// it starts as boolean
-static unsigned int  buf_size=256;	// measured in 1KBs
+static unsigned int  dao_blocks=0;
 static unsigned char formats[260],disc_info[32];
 
-extern int	dvd_compat,test_write,no_reload,mmc_profile,_argc,wrvfy;
+extern "C" {
+extern int	dvd_compat,test_write,no_reload,mmc_profile,_argc,
+		wrvfy,no_opc;
 extern double	speed_factor;
 extern char	*ioctl_device,**_argv;
+}
 
 extern "C"
 int fumount (int fd)
@@ -58,17 +68,7 @@ int media_reload (char *name=NULL,struct stat *sb=NULL)
 {   if (name==NULL)
     {	Scsi_Command cmd(ioctl_handle);
 
-	while (1)	// Pioneer DVR-x06 needs this...
-	{   cmd[0] = 0x1B;	// START/STOP UNIT
-	    cmd[1] = 0x1;	// "IMMED"
-	    cmd[4] = 0;		// "Stop"
-	    cmd[5] = 0;
-	    if (cmd.transport() == 0x20407)	// "OP IN PROGRESS"
-	    {	poll (NULL,0,333);
-		continue;
-	    }
-	    break;
-	}
+	pioneer_stop (cmd);
 
 #if defined(RELOAD_NEVER_NEEDED)
 #undef RELOAD_NEVER_NEEDED
@@ -91,8 +91,9 @@ int media_reload (char *name=NULL,struct stat *sb=NULL)
 	if ((n=fcntl (ioctl_fd,F_GETFD))<0) n=0;
 	fcntl (ioctl_fd,F_SETFD,n&~FD_CLOEXEC);
 
-	sprintf (str,"%d",ioctl_fd);
-	execlp(_argv[0],"-reload",str,ioctl_device,NULL);
+	sprintf (str,"%ld",ioctl_fd);
+	execlp(_argv[0],no_reload<0?"-eject":"-reload",
+			str,ioctl_device,(void *)NULL);
     }
     else
     {
@@ -126,7 +127,7 @@ int media_reload (char *name=NULL,struct stat *sb=NULL)
 #if defined(__sun) || defined(sun)
 	    else if (volmgt_running())
 	    {	setuid(getuid());
-		execl("/usr/bin/volrmmount","volrmmount","-i",name,NULL);
+		execl("/usr/bin/volrmmount","volrmmount","-i",name,(void*)NULL);
 		return 0;	// not normally reached
 	    }
 #endif
@@ -297,6 +298,8 @@ static unsigned int get_2k_capacity (Scsi_Command &cmd)
 
 	case 0x11:		// DVD-R
 	case 0x14:		// DVD-RW Sequential
+	case 0x15:		// DVD-R Dual Layer Sequential
+	case 0x16:		// DVD-R Dual Layer Jump
 	case 0x1B:		// DVD+R
 	case 0x2B:		// DVD+R Double Layer
 	    cmd[0] = 0x52;	// READ TRACK INFORMATION
@@ -337,6 +340,24 @@ off64_t get_capacity (void *fd)
   return (off64_t)get_2k_capacity(cmd)*2048;
 }
 
+extern "C"
+double get_buffer_stats (void *fd)
+{ Scsi_Command		cmd(fd);
+  unsigned char		bcap[12];
+  unsigned int		bsize,bfree;
+
+    cmd[0] = 0x5C;              // READ BUFFER CAPACITY
+    cmd[8] = sizeof(bcap);
+    cmd[9] = 0;
+    if (cmd.transport (READ,bcap,sizeof(bcap)))
+	return 0.0;
+
+    bsize = bcap[4]<<24|bcap[5]<<16|bcap[6]<<8|bcap[7];
+    bfree = bcap[8]<<24|bcap[9]<<16|bcap[10]<<8|bcap[11];
+
+  return (1.0 - (double)bfree/(double)bsize);
+}
+
 ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 { Scsi_Command		cmd(ioctl_handle);	/* screw first argument */
   unsigned char		bcap[12];
@@ -369,11 +390,14 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
     { static unsigned int first_wr_addr=0;
 
 	if (!media_written)
-	{   sigs_mask (SIG_BLOCK); first_wr_addr = lba;	}
+	{   sigs_mask (SIGS_BLOCK); first_wr_addr = lba;	}
 	else if (lba >= (first_wr_addr+buf_size/2)) // measured in 2KBs!
-	{   sigs_mask (SIG_UNBLOCK);			}
+	{   sigs_mask (SIGS_UNBLOCK);			}
     }
 
+    if (dao_blocks!=0 && (lba+nbl)>dao_blocks)
+	nbl = dao_blocks-lba;
+
     while (1)
     {	cmd[0] = wrvfy?0x2E:0x2A;	// WRITE [AND VERIFY] (10)
 	cmd[2] = (lba>>24)&0xff;	// Logical Block Addrss
@@ -488,13 +512,22 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 		    msecs /= velocity;
 		}
 		else if (!retries++)	continue;
+		// Pioneer units seem to return bogus values at high
+		// velocities now and then... I reserve for 4 ECC
+		// blocks, which is ~6 ms at 16x. But note that some
+		// OSes, Linux 2.4 among then, will nap for 10 or 20.
+		msecs = (4*32*1000)/velocity;
 	    }
 
-	    if (retries > (dao_toggle?192:96))
+	    // Retry values were increased because high-speed units
+	    // start recordings at only portion of velocity and it
+	    // takes more retries to write lead-in... Another common
+	    // reason for stucks are recalibrations at zone edges...
+	    if (retries > (dao_toggle?768:192))
 		fprintf (stderr,":-? the LUN appears to be stuck "
-				"writing LBA=%xh, retry in %dms\n",
+				"writing LBA=%xh, keep retrying in %dms\n",
 				lba,msecs),
-		retries=0, msecs--;
+		retries=1;
 
 	    if (msecs>=0)
 	    {	poll (NULL,0,msecs);
@@ -503,6 +536,21 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 
 	    lba |= 0x80000000;		// signal insane bfree...
 	}
+	else if (lba==0 && errcode==0x20401)	// "IN PROCESS OF BECOMING READY"
+	{   if ((handle_events(cmd) & (1<<6)))
+		continue;
+	    if (++retries > 3)
+	    {	fprintf (stderr,":-! the LUN is still in process of becoming ready, "
+		    		    "retrying in 5 secs...\n"),
+		retries=0;
+		sigs_mask (SIGS_UNBLOCK);
+		poll (NULL,0,5000);
+	    }
+	    else
+	    	sigs_mask (SIGS_UNBLOCK),
+		poll (NULL,0,stdnap);
+	    continue;
+	}
 	else if (errcode==0x20404 ||	// "FORMAT IN PROGRESS"
 		 errcode==0x20407 ||	// "OPERATION IN PROGRESS"
 		 errcode==0x52C00)	// "COMMAND SEQUENCE ERROR"
@@ -517,7 +565,7 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 	    	{   fprintf (stderr,":-! the LUN appears to be stuck at %xh, "
 		    		    "retrying in 5 secs...\n",lba),
 		    retries=0;
-		    sigs_mask (SIG_UNBLOCK);
+		    sigs_mask (SIGS_UNBLOCK);
 		    poll (NULL,0,5000);
 		}
 		else if (errcode==0x52C00)
@@ -895,12 +943,17 @@ static int set_speed_B6h (Scsi_Command &cmd,unsigned int dvddash,
 		if (v<minv) minv=v;
 	    } // minv will be treated as 1x
 
-	    // 2.4x is like 1x for DVD+, but it turned out that there're
-	    // units, most notably "DVDRW IDE1004," burning DVD+ at
-	    // lower speed, so I have to watch out for it...
-	    if (!dvddash && minv>(2*ONEX))
+	    // New generation units offer minimum velocities other than 1x
+	    if      (minv>(15*ONEX))	speed_factor /= 16.0;
+	    else if (minv>(11*ONEX))	speed_factor /= 12.0;
+	    else if (minv>(7*ONEX))	speed_factor /= 8.0;
+	    else if (minv>(5*ONEX))	speed_factor /= 6.0;
+	    else if (minv>(3*ONEX))	speed_factor /= 4.0;	
+	    else if (!dvddash && minv>(2*ONEX))
+		// 2.4x is like 1x for DVD+, but it turned out that there're
+		// units, most notably "DVDRW IDE1004," burning DVD+ at
+		// lower speed, so I have to watch out for it...
 		speed_factor /= 2.4;
-	    // some units, e.g. NEC ND-2500, offer 2x as minimal speed...
 	    else if (minv>=(2*ONEX))
 		speed_factor /= 2.0;
 
@@ -965,9 +1018,16 @@ static int set_speed_B6h (Scsi_Command &cmd,unsigned int dvddash,
 	{   if (targetv!=velocity)
 	    {	if (errp==0)	// check Page 2A for speed then...
 		{   errp=-1; continue;   }
-		fprintf (stderr,":-( Failed to change write speed: "
-				"%d->%d\n",velocity,targetv),
-		exit (FATAL_START(EINVAL));
+		if (velocity==0 || velocity>=ONEX)
+		{   fprintf(stderr,":-? Failed to change write speed: %d->%d\n",
+					(int)velocity,targetv);
+		    if (velocity<targetv)	velocity=targetv;
+		}
+		else
+		{   fprintf(stderr,":-( Failed to change write speed: %d->%d\n",
+					(int)velocity,targetv);
+		    exit (FATAL_START(EINVAL));
+		}
 	    }
 	    break;
 	}
@@ -1038,7 +1098,7 @@ static int set_speed_BBh (Scsi_Command &cmd,unsigned char *page2A)
 	else if (targetv)
 	{   if (targetv!=velocity)
 		fprintf (stderr,":-( Failed to change write speed: %d->%d\n",
-				velocity,targetv),
+				(int)velocity,targetv),
 		exit (FATAL_START(EINVAL));
 	    break;
 	}
@@ -1077,6 +1137,7 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 	    break;
 	case 0x11:	// DVD-R Sequential
 	case 0x14:	// DVD-RW Sequential
+	case 0x15:	// DVD-R Dual Layer Sequential
 	    p32 = 0xC0;
 	    if (next_track==1) do
 	    {	if (dvd_compat >= (profile==0x14?2:256))
@@ -1107,8 +1168,18 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 	    } while (0);
 	
 	    if (is_dao)		p32 |=2;	// DAO
-	    if (dvd_compat)	p32 &= 0x3F;	// Single-session
 
+	    if (profile==0x15 || // DVD-R DL Seq has no notion of multi-session
+		dvd_compat)	p32 &= 0x3F;	// Single-session
+
+	    if (test_write)
+		p32 |= 0x10;	// Test Write for debugging purposes
+
+	    dvddash=1;
+	    break;
+	case 0x16:	// DVD-R Dual Layer Jump
+	    p32 = 0xC0;
+	    p32 |= 4;	// Jump
 	    if (test_write)
 		p32 |= 0x10;	// Test Write for debugging purposes
 
@@ -1130,11 +1201,11 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
     unsigned char * const p = page2A + hlen;
 
     if (plen<(hlen+14) || p[1]<(14-2))	// no "Buffer Size Supported"
-	buf_size = 256;			// bogus value
+	buf_size = 512;			// bogus value
     else
     	buf_size = (p[12]<<8|p[13]);
-    if (buf_size<256)			// bogus value
-	buf_size = 256;
+    if (buf_size<512)			// bogus value
+	buf_size = 512;
 
     // GET PERFORMANCE/SET STREAMING are listed as mandatory, so that
     // this is actually the most likely path...
@@ -1186,7 +1257,7 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 			"-speed option is ignored\n");
 opc:
 
-    if (!dvddash) return 0;
+    if (!dvddash || test_write || no_opc) return 0;
 
     // See if OPC is required...
     cmd[0] = 0x51;	// READ DISC INFORMATION
@@ -1202,17 +1273,20 @@ opc:
 	cmd[9] = 0;
 	cmd.timeout(120);	// NEC units can be slooo...w
 	if ((err=cmd.transport())) do
-	{   if (err==0x52000)	// "INVALID COMMAND"
-		break;		// LD-F321 doesn't support SEND OPC?
-	    if (err==0x52700)	// "WRITE PROTECTED"
-		break;		// Plextor PX-712A?
-	    if (err==0x17301)	// "POWER CALIBRATION AREA ALMOST FULL"
+	{   if (err==0x17301)	// "POWER CALIBRATION AREA ALMOST FULL"
 	    {	fprintf (stderr,":-! WARNING: Power Calibration Area "
 	    			    "is almost full\n");
 		break;
 	    }
-	    sperror ("PERFORM OPC",err),
-	    exit(FATAL_START(errno));
+	    sperror ("PERFORM OPC",err);
+	    if ((err&0x0FF00)==0x07300)	exit(FATAL_START(errno));
+	    /*
+	     * Th rest of errors are ignored, most notably observed:
+	     * 0x52000, "INVALID COMMAND"		Panasonic LD-F321
+	     * 0x52700, "WRITE PROTECTED"		Plextor PX-712A
+	     * 0x52C00, "COMMAND SEQUENCE ERROR"	Plextor PX-716UF
+	     * 0x53002, "MEDIA NOT SUPPORTED"		Toshiba TS-H542A
+	     */
 	} while (0);
     }
 
@@ -1286,12 +1360,13 @@ static int minus_r_reserve_track (Scsi_Command &cmd,off64_t size)
   unsigned int blocks;
 
     blocks = size/2048;
-    blocks += 15, blocks &= ~15;
+    if (is_dao)	dao_blocks = blocks;
+    else	blocks += 15, blocks &= ~15;
 
     fprintf (stderr,"%s: reserving %u blocks",ioctl_device,blocks);
     if (is_dao && blocks<380000)
 	fprintf (stderr,"\b, warning for short DAO recording"),
-	poll (NULL,0,3000);
+	poll (NULL,0,5000);
     fprintf (stderr,"\n");
 			
 
@@ -1366,7 +1441,7 @@ static void plus_r_dl_split (Scsi_Command &cmd,off64_t size)
 
 }
 
-static int flush_cache (Scsi_Command &cmd)
+static int flush_cache (Scsi_Command &cmd, int even_sync=1)
 { int err;
 
     cmd[0] = 0x35;		// FLUSH CACHE
@@ -1377,15 +1452,16 @@ static int flush_cache (Scsi_Command &cmd)
     else
 	sperror ("FLUSH CACHE",err);
 
-#if 1	// Pioneer apparently needs this, non-IMMED FLUSH that is...
-    cmd[0] = 0x35;		// FLUSH CACHE
-    cmd[9] = 0;
-    if (is_dao) cmd.timeout (15*60);
-    if ((err=cmd.transport()))
-    {	sperror ("SYNCHRONOUS FLUSH CACHE",err);
-	return err;
+    if (even_sync)		// Pioneer apparently needs this,
+    {				// non-IMMED FLUSH that is...
+	cmd[0] = 0x35;		// FLUSH CACHE
+	cmd[9] = 0;
+	if (is_dao) cmd.timeout (15*60);
+	if ((err=cmd.transport()))
+	{   sperror ("SYNCHRONOUS FLUSH CACHE",err);
+	    return err;
+	}
     }
-#endif
 
   return 0;
 }
@@ -1401,6 +1477,8 @@ void no_r_finalize ()
 	fprintf (stderr,"%s: flushing cache\n",ioctl_device);
 	if (flush_cache (cmd))	break;
 
+	pioneer_stop (cmd);
+
 	media_written = 0;
 	errno = 0;
     }
@@ -1409,7 +1487,7 @@ void no_r_finalize ()
 
 extern "C"
 void plus_rw_finalize ()
-{   sigs_mask(SIG_BLOCK);
+{   sigs_mask(SIGS_BLOCK);
     while (media_written)
     { Scsi_Command cmd(ioctl_handle);
       int err;
@@ -1451,7 +1529,7 @@ void plus_rw_finalize ()
 
 extern "C"
 void plus_r_finalize ()
-{   sigs_mask(SIG_BLOCK);
+{   sigs_mask(SIGS_BLOCK);
     while (media_written)
     { Scsi_Command cmd(ioctl_handle);
       int          mode,err;
@@ -1480,12 +1558,20 @@ void plus_r_finalize ()
 	    mode = 0x02,
 	    fprintf (stderr,"%s: closing session\n",ioctl_device);
 
-	cmd[0] = 0x5B;		// CLOSE TRACK/SESSION
-	cmd[1] = 0x01;		// "IMMED"
-	cmd[2] = mode;		// "Close session"
-	cmd[9] = 0;
-	if ((err=cmd.transport()))
-	    sperror ("CLOSE SESSION",err);
+	while (1)	// Pioneer DVR-K04RA
+	{   cmd[0] = 0x5B;		// CLOSE TRACK/SESSION
+	    cmd[1] = 0x01;		// "IMMED"
+	    cmd[2] = mode;		// "Close session"
+	    cmd[9] = 0;
+	    err = cmd.transport();
+	    if (err == 0x20407)		// "OP IN PROGRESS"
+	    {	poll (NULL,0,stdnap);
+		continue;
+	    }
+	    else if (err)
+		sperror (dvd_compat?"CLOSE DISC":"CLOSE SESSION",err);
+	    break;
+	}
 
 	if (wait_for_unit (cmd)) break;
  
@@ -1500,7 +1586,7 @@ void plus_r_finalize ()
 
 extern "C"
 void minus_rw_finalize ()
-{   sigs_mask(SIG_BLOCK);
+{   sigs_mask(SIGS_BLOCK);
     while (media_written)
     { Scsi_Command cmd(ioctl_handle);
       int err;
@@ -1537,13 +1623,13 @@ void minus_rw_finalize ()
 
 extern "C"
 void minus_r_finalize ()
-{   sigs_mask(SIG_BLOCK);
+{   sigs_mask(SIGS_BLOCK);
     while (media_written)
     { Scsi_Command cmd(ioctl_handle);
       int err;
 
 	fprintf (stderr,"%s: flushing cache\n",ioctl_device);
-	if (flush_cache(cmd))	break;
+	if (flush_cache(cmd,!is_dao))	break;
 
 	if (!is_dao)
 	{   fprintf (stderr,"%s: updating RMA\n",ioctl_device);
@@ -1560,12 +1646,21 @@ void minus_r_finalize ()
 
 	    fprintf (stderr,"%s: closing %s\n",ioctl_device,
 			    dvd_compat?"disc":"session");
-	    cmd[0] = 0x5B;	// CLOSE TRACK/SESSION
-	    cmd[1] = 0x01;	// "IMMED"
-	    cmd[2] = 0x02;	// "Close Session"
-	    cmd[9] = 0;
-	    if ((err=cmd.transport()))
-		sperror (dvd_compat?"CLOSE DISC":"CLOSE SESSION",err);
+
+	    while (1)	// Pioneer DVR-K04RA
+	    {	cmd[0] = 0x5B;	// CLOSE TRACK/SESSION
+		cmd[1] = 0x01;	// "IMMED"
+		cmd[2] = 0x02;	// "Close Session"
+		cmd[9] = 0;
+		err = cmd.transport();
+		if (err == 0x20407)		// "OP IN PROGRESS"
+		{   poll (NULL,0,stdnap);
+		    continue;
+		}
+		else if (err)
+		    sperror (dvd_compat?"CLOSE DISC":"CLOSE SESSION",err);
+		break;
+	    }
 
 	    if (wait_for_unit (cmd)) break;
 	}
@@ -1581,7 +1676,7 @@ void minus_r_finalize ()
 
 extern "C"
 void ram_reload ()
-{   sigs_mask(SIG_BLOCK);
+{   sigs_mask(SIGS_BLOCK);
     if (media_written && do_reload) media_reload();
     _exit (errno);
 }
@@ -1591,27 +1686,6 @@ void ram_reload ()
 // It's invoked right before first write and if necessary/applicable
 // prepares Page 05, reserves track, sets up atexit and signal handlers.
 //
-#define sigit(SIG,f)	do {	\
-    sigaction (SIG,NULL,&sa);	\
-    sa.sa_mask = mask;		\
-    sa.sa_handler = f;		\
-    sa.sa_flags &= ~SA_NODEFER;	\
-    sa.sa_flags |= SA_RESETHAND;\
-    sigaction (SIG,&sa,NULL);	\
-} while (0)
-static void atsignals(void(*func)())
-{ void           (*f)(int)=(void(*)(int))func;
-  sigset_t         mask;
-  struct sigaction sa;
-
-    sigemptyset (&mask);
-    sigaddset (&mask,SIGHUP),	sigaddset (&mask,SIGINT),
-    sigaddset (&mask,SIGTERM),	sigaddset (&mask,SIGPIPE);
-
-    sigit (SIGHUP,f);	sigit (SIGINT,f);
-    sigit (SIGTERM,f);	sigit (SIGPIPE,f);
-}
-#undef sigit
 
 typedef ssize_t (*pwrite64_t)(int,const void *,size_t,off64_t);
 
@@ -1673,7 +1747,29 @@ pwrite64_t poor_mans_setup (void *fd,off64_t leadout)
 	    break;
 	case 0x11:	// DVD-R Sequential
 	case 0x14:	// DVD-RW Sequential
+	case 0x15:	// DVD-R Dual Layer Sequential
 	    plusminus_pages_setup (cmd,profile);
+	    if ((profile==0x15) && leadout)
+	    {	unsigned char dvd_10[4+16];
+		unsigned int layer_size,data_size=leadout/2048;
+
+		cmd[0] = 0xAD;	// READ DVD STRUCTURE
+		cmd[7] = 0x10;
+		cmd[9] = sizeof(dvd_10);
+		cmd[11] = 0;
+		if ((err=cmd.transport(READ,dvd_10,sizeof(dvd_10))))
+		    sperror("READ DVD STRUCTURE #10h",err),
+		    exit(FATAL_START(errno));
+
+		layer_size  = dvd_10[4+13]<<16,
+		layer_size |= dvd_10[4+14]<<8,
+		layer_size |= dvd_10[4+15];
+
+		if (data_size <= layer_size)
+		    fprintf (stderr,":-( more than 50%% of space will be *wasted*!\n"
+				    "    use single layer media for this recording\n"),
+		    exit(FATAL_START(EMEDIUMTYPE)); 
+	    }
 	    if (is_dao && leadout)
 		minus_r_reserve_track(cmd,leadout);
 	    atexit (minus_r_finalize);
@@ -1689,6 +1785,8 @@ pwrite64_t poor_mans_setup (void *fd,off64_t leadout)
 	    atexit (ram_reload);
 	    atsignals (ram_reload);
 	    break;
+	case 0x16:	// DVD-R Dual Layer Jump not supported for reasons
+			// discussed at web-page...
 	default:
 	    fprintf (stderr,":-( mounted media[%X] is not supported\n",profile),
 	    exit(FATAL_START(EMEDIUMTYPE));
@@ -1738,7 +1836,8 @@ int poor_man_rewritable (void *fd, void *buf)
 
     if (profile!=0x13 &&	// not DVD-RW Restricted Overwrite
 	profile!=0x12 &&	// nor DVD-RAM
-	profile!=0x1A)		// nor DVD+RW
+	profile!=0x1A &&	// nor DVD+RW
+	profile!=0x2A)		// nor DVD+RW Double Layer
 	return 0;
 
     if (profile==0x13)		// DVD-RW Restricted Overwrite
@@ -1748,7 +1847,8 @@ int poor_man_rewritable (void *fd, void *buf)
 	minus_rw_finalize ();	// with do_reload==0!
 	do_reload=1;
     }
-    else if (profile==0x1A)	// DVD+RW
+    else if (profile==0x1A ||	// DVD+RW
+	     profile==0x2A)	// DVD+RW Double Layer
     {	fprintf (stderr,"%s: flushing cache\n",ioctl_device);
 	if (flush_cache(cmd))	exit (errno);
     }
